@@ -25,7 +25,9 @@ __all__ = [
 import logging
 
 from systemimage.helpers import last_update_date
+from systemimage.reboot import factory_reset
 from systemimage.state import State
+from unittest.mock import patch
 
 log = logging.getLogger('systemimage')
 
@@ -78,7 +80,7 @@ class Mediator:
         self._update = None
         self._callback = callback
 
-    def __repr__(self):
+    def __repr__(self): # pragma: no cover
         return '<Mediator at 0x{:x} | State at 0x{:x}>'.format(
             id(self), id(self._state))
 
@@ -113,14 +115,13 @@ class Mediator:
     def download(self):
         """Download the available update."""
         # We only want callback progress during the actual download.
-        callback = self._state.downloader.callback
-        try:
-            self._state.downloader.callback = self._callback
+        with patch.object(self._state.downloader, 'callback', self._callback):
             self._state.run_until('reboot')
-        finally:
-            self._state.downloader.callback = callback
 
     def reboot(self):
         """Issue the reboot."""
         # Transition through all remaining states.
         list(self._state)
+
+    def factory_reset(self):
+        factory_reset()
