@@ -19,16 +19,31 @@ __all__ = [
     'Canceled',
     'DBusDownloadManager',
     'DuplicateDestinationError',
+    'Record',
     ]
 
 
+import os
 import dbus
 import logging
 
+from collections import namedtuple
+from contextlib import suppress
 from io import StringIO
 from pprint import pformat
 from systemimage.config import config
 from systemimage.reactor import Reactor
+from systemimage.settings import Settings
+
+# The systemimage.testing module will not be available on installed systems
+# unless the system-image-dev binary package is installed, which is not usually
+# the case.  Disable _print() debugging in that case.
+def _print(*args, **kws):
+    with suppress(ImportError):
+        # We must import this here to avoid circular imports.
+        from systemimage.testing.helpers import debug
+        with debug(check_flag=True) as ddlog:
+            ddlog(*args, **kws)
 
 
 # Parameterized for testing purposes.
@@ -61,6 +76,19 @@ class DuplicateDestinationError(Exception):
         return '\n' + pformat(self.duplicates, indent=4, width=79)
 
 
+# A namedtuple is convenient here since we want to access items by their
+# attribute names.  However, we also want to allow for the checksum to default
+# to the empty string.  We do this by creating a prototypical record type and
+# using _replace() to replace non-default values.  See the namedtuple
+# documentation for details.
+_Record = namedtuple('Record', 'url destination checksum')('', '', '')
+_RecordType = type(_Record)
+
+def Record(url, destination, checksum=''):
+    return _Record._replace(
+        url=url, destination=destination, checksum=checksum)
+
+
 class DownloadReactor(Reactor):
     def __init__(self, bus, callback=None, pausable=False):
         super().__init__(bus)
@@ -68,6 +96,8 @@ class DownloadReactor(Reactor):
         self._pausable = pausable
         self.error = None
         self.canceled = False
+        self.received = 0
+        self.total = 0
         self.react_to('canceled')
         self.react_to('error')
         self.react_to('finished')
@@ -76,27 +106,23 @@ class DownloadReactor(Reactor):
         self.react_to('resumed')
         self.react_to('started')
 
-    def _print(self, *args, **kws):
-        ## from systemimage.testing.helpers import debug
-        ## with debug() as ddlog:
-        ##     ddlog(*args, **kws)
-        pass
-
     def _do_started(self, signal, path, started):
-        self._print('STARTED:', started)
+        _print('STARTED:', started)
 
     def _do_finished(self, signal, path, local_paths):
-        self._print('FINISHED:', local_paths)
+        _print('FINISHED:', local_paths)
         self.quit()
 
     def _do_error(self, signal, path, error_message):
-        self._print('ERROR:', error_message)
+        _print('ERROR:', error_message)
         log.error(error_message)
         self.error = error_message
         self.quit()
 
     def _do_progress(self, signal, path, received, total):
-        self._print('PROGRESS:', received, total)
+        self.received = received
+        self.total = total
+        _print('PROGRESS:', received, total)
         if self._callback is not None:
             # Be defensive, so yes, use a bare except.  If an exception occurs
             # in the callback, log it, but continue onward.
@@ -109,25 +135,28 @@ class DownloadReactor(Reactor):
         # Why would we get this signal if it *wasn't* canceled?  Anyway,
         # this'll be a D-Bus data type so converted it to a vanilla Python
         # boolean.
-        self._print('CANCELED:', canceled)
+        _print('CANCELED:', canceled)
         self.canceled = bool(canceled)
         self.quit()
 
     def _do_paused(self, signal, path, paused):
-        self._print('PAUSE:', paused, self._pausable)
-        if self._pausable and config.dbus_service is not None:
+        _print('PAUSE:', paused, self._pausable)
+        send_paused = self._pausable and config.dbus_service is not None
+        if send_paused:                             # pragma: no branch
             # We could plumb through the `service` object from service.py (the
             # main entry point for system-image-dbus, but that's actually a
             # bit of a pain, so do the expedient thing and grab the interface
             # here.
-            config.dbus_service.UpdatePaused(0)
+            percentage = (int(self.received / self.total * 100.0)
+                          if self.total > 0 else 0)
+            config.dbus_service.UpdatePaused(percentage)
 
     def _do_resumed(self, signal, path, resumed):
-        self._print('RESUME:', resumed)
+        _print('RESUME:', resumed)
         # There currently is no UpdateResumed() signal.
 
     def _default(self, *args, **kws):
-        self._print('SIGNAL:', args, kws)
+        _print('SIGNAL:', args, kws)                # pragma: no cover
 
 
 class DBusDownloadManager:
@@ -143,7 +172,7 @@ class DBusDownloadManager:
         self._queued_cancel = False
         self.callback = callback
 
-    def __repr__(self):
+    def __repr__(self): # pragma: no cover
         return '<DBusDownloadManager at 0x{:x}>'.format(id(self))
 
     def get_files(self, downloads, *, pausable=False):
@@ -160,9 +189,12 @@ class DBusDownloadManager:
         all-or-nothing function.  Either you get all the requested files
         or none of them.
 
-        :param downloads: A list of 2-tuples where the first item is the url to
-            download, and the second item is the destination file.
-        :type downloads: List of 2-tuples.
+        :params downloads: A list of `download records`, each of which may
+            either be a 2-tuple where the first item is the url to download,
+            and the second item is the destination file, or an instance of a
+            `Record` namedtuple with attributes `url`, `destination`, and
+            `checksum`.  The checksum may be the empty string.
+        :type downloads: List of 2-tuples or `Record`s.
         :param pausable: A flag specifying whether this download can be paused
             or not.  In general, data file downloads are pausable, but
             preliminary downloads are not.
@@ -179,35 +211,67 @@ class DBusDownloadManager:
         if len(downloads) == 0:
             # Nothing to download.  See LP: #1245597.
             return
-        destinations = set(dst for url, dst in downloads)
+        # Convert the downloads items to download records.
+        records = [item if isinstance(item, _RecordType) else Record(*item)
+                   for item in downloads]
+        destinations = set(record.destination for record in records)
+        # Check for duplicate destinations, specifically for a local file path
+        # coming from two different sources.  It's okay if there are duplicate
+        # destination records in the download request, but each of those must
+        # be specified by the same source url and have the same checksum.
+        #
+        # An easy quick check just asks if the set of destinations is smaller
+        # than the total number of requested downloads.  It can't be larger.
+        # If it *is* smaller, then there are some duplicates, however the
+        # duplicates may be legitimate, so look at the details.
+        #
+        # Note though that we cannot pass duplicates destinations to udm,
+        # so we have to filter out legitimate duplicates.  That's fine since
+        # they really are pointing to the same file, and will end up in the
+        # destination location.
         if len(destinations) < len(downloads):
-            # Spend some extra effort to provide reasonable exception details.
-            reverse = {}
-            for url, dst in downloads:
-                reverse.setdefault(dst, []).append(url)
+            by_destination = dict()
+            unique_downloads = set()
+            for record in records:
+                by_destination.setdefault(record.destination, set()).add(
+                    record)
+                unique_downloads.add(record)
             duplicates = []
-            for dst, urls in reverse.items():
-                if len(urls) > 1:
-                    duplicates.append((dst, urls))
-            raise DuplicateDestinationError(sorted(duplicates))
+            for dst, seen in by_destination.items():
+                if len(seen) > 1:
+                    # Tuples will look better in the pretty-printed output.
+                    duplicates.append(
+                        (dst, sorted(tuple(dup) for dup in seen)))
+            if len(duplicates) > 0:
+                raise DuplicateDestinationError(sorted(duplicates))
+            # Uniquify the downloads.
+            records = list(unique_downloads)
         bus = dbus.SystemBus()
         service = bus.get_object(DOWNLOADER_INTERFACE, '/')
         iface = dbus.Interface(service, MANAGER_INTERFACE)
         # Better logging of the requested downloads.
         fp = StringIO()
         print('[0x{:x}] Requesting group download:'.format(id(self)), file=fp)
-        for url, dst in downloads:
-            print('\t{} -> {}'.format(url, dst), file=fp)
+        for record in records:
+            if record.checksum == '':
+                print('\t{} -> {}'.format(*record[:2]), file=fp)
+            else:
+                print('\t{} [{}] -> {}'.format(*record), file=fp)
         log.info('{}'.format(fp.getvalue()))
         object_path = iface.createDownloadGroup(
-            [(url, dst, '') for url, dst in downloads],
-            '',           # No hashes yet.
+            records,
+            'sha256',
             False,        # Don't allow GSM yet.
             # https://bugs.freedesktop.org/show_bug.cgi?id=55594
             dbus.Dictionary(signature='sv'),
             _headers())
         download = bus.get_object(OBJECT_NAME, object_path)
         self._iface = dbus.Interface(download, OBJECT_INTERFACE)
+        # Are GSM downloads allowed?  Yes, except if auto_download is set to 1
+        # (i.e. wifi-only).
+        allow_gsm = Settings().get('auto_download') != '1'
+        DBusDownloadManager._set_gsm(self._iface, allow_gsm=allow_gsm)
+        # Start the download.
         reactor = DownloadReactor(bus, self.callback, pausable)
         reactor.schedule(self._iface.start)
         log.info('[0x{:x}] Running group download reactor', id(self))
@@ -228,6 +292,15 @@ class DBusDownloadManager:
             raise Canceled
         if reactor.timed_out:
             raise TimeoutError
+        # For sanity.
+        for record in records:
+            assert os.path.exists(record.destination), (
+                'Missing destination: {}'.format(record))
+
+    @staticmethod
+    def _set_gsm(iface, *, allow_gsm):
+        # This is a separate method for easier testing via mocks.
+        iface.allowGSMDownload(allow_gsm)
 
     def cancel(self):
         """Cancel any current downloads."""
@@ -244,10 +317,10 @@ class DBusDownloadManager:
 
     def pause(self):
         """Pause the download, but only if one is in progress."""
-        if self._iface is not None:
+        if self._iface is not None:                 # pragma: no branch
             self._iface.pause()
 
     def resume(self):
         """Resume the download, but only if one is in progress."""
-        if self._iface is not None:
+        if self._iface is not None:                 # pragma: no branch
             self._iface.resume()
