@@ -71,7 +71,7 @@ def stop_system_image(controller):
 
 def _find_udm_process():
     for process in psutil.process_iter():
-        cmdline = SPACE.join(process.cmdline)
+        cmdline = SPACE.join(process.cmdline())
         if 'ubuntu-download-manager' in cmdline and '-stoppable' in cmdline:
             return process
     return None
@@ -112,7 +112,7 @@ DLSERVICE = '/usr/bin/ubuntu-download-manager'
 
 SERVICES = [
    ('com.canonical.SystemImage',
-    '{python} -m systemimage.service -C {self.ini_path} --testing {self.mode}',
+    '{python} -m {self.MODULE} -C {self.ini_path} --testing {self.mode}',
     start_system_image,
     stop_system_image,
    ),
@@ -128,7 +128,9 @@ SERVICES = [
 class Controller:
     """Start and stop D-Bus service under test."""
 
-    def __init__(self, logfile=None):
+    MODULE = 'systemimage.testing.service'
+
+    def __init__(self, logfile=None, loglevel='info'):
         # Non-public.
         self._stack = ExitStack()
         self._stoppers = []
@@ -158,16 +160,14 @@ class Controller:
         template = resource_bytes(
             'systemimage.tests.data', 'config_03.ini').decode('utf-8')
         with open(self.ini_path, 'w', encoding='utf-8') as fp:
-            print(template.format(tmpdir=ini_tmpdir, vardir=ini_vardir,
-                                  logfile=ini_logfile),
+            print(template.format(tmpdir=ini_tmpdir,
+                                  vardir=ini_vardir,
+                                  logfile=ini_logfile,
+                                  loglevel=loglevel),
                   file=fp)
 
     def _configure_services(self):
-        # If the dbus-daemon is already running, kill all the children.
-        if self.daemon_pid is not None:
-            for stopper in self._stoppers:
-                stopper(self)
-        del self._stoppers[:]
+        self.stop_children()
         # Now we have to set up the .service files.  We use the Python
         # executable used to run the tests, executing the entry point as would
         # happen in a deployed script or virtualenv.
@@ -244,9 +244,15 @@ class Controller:
             self._stack.close()
             raise
 
+    def stop_children(self):
+        # If the dbus-daemon is already running, kill all the children.
+        if self.daemon_pid is not None:
+            for stopper in self._stoppers:
+                stopper(self)
+        del self._stoppers[:]
+
     def _kill(self, pid):
-        for stopper in self._stoppers:
-            stopper(self)
+        self.stop_children()
         process = psutil.Process(pid)
         process.terminate()
         process.wait(60)
