@@ -27,8 +27,8 @@ from dbus.service import method, signal
 from gi.repository import GLib
 from systemimage.api import Mediator
 from systemimage.config import config
-from systemimage.dbus import Service
-from systemimage.helpers import makedirs, safe_remove
+from systemimage.dbus import Service, log_and_exit
+from systemimage.helpers import MiB, makedirs, safe_remove, version_detail
 from unittest.mock import patch
 
 
@@ -62,6 +62,7 @@ def instrument(config, stack):
 class _LiveTestableService(Service):
     """For testing purposes only."""
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def Reset(self):
         self._api = Mediator()
@@ -77,6 +78,7 @@ class _LiveTestableService(Service):
         del config.build_number
         safe_remove(config.system.settings_db)
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def TearDown(self):
         # Like CancelUpdate() except it sends a different signal that's only
@@ -84,6 +86,7 @@ class _LiveTestableService(Service):
         self._api.cancel()
         self.TornDown()
 
+    @log_and_exit
     @signal('com.canonical.SystemImage')
     def TornDown(self):
         pass
@@ -106,33 +109,27 @@ class _UpdateAutoSuccess(Service):
         self._percentage = 0
         self._rebootable = False
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def Reset(self):
         self._reset()
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def CheckForUpdate(self):
         if self._failure_count > 0:
             self._reset()
         GLib.timeout_add_seconds(3, self._send_status)
 
+    @log_and_exit
     def _send_status(self):
-        self.UpdateAvailableStatus(
-            True, self._auto_download, '42', 1337 * 1024 * 1024,
-            '1983-09-13T12:13:14',
-            #[
-            #{'description': 'Ubuntu Edge support',
-            # 'description-en_GB': 'change the background colour',
-            # 'description-fr': "Support d'Ubuntu Edge",
-            #},
-            #{'description':
-            # 'Flipped container with 200% boot speed improvement',
-            #"}],
-            '')
-        if (    self._auto_download and
-                not self._downloading and
-                not self._rebootable):
+        if self._auto_download:
             self._downloading = True
+        self.UpdateAvailableStatus(
+            True, self._downloading, '42', 1337 * MiB,
+            '1983-09-13T12:13:14',
+            '')
+        if self._downloading and not self._rebootable:
             self.UpdateProgress(0, 50.0)
             GLib.timeout_add(500, self._send_more_status)
         if self._paused:
@@ -141,6 +138,7 @@ class _UpdateAutoSuccess(Service):
             self.UpdateDownloaded()
         return False
 
+    @log_and_exit
     def _send_more_status(self):
         if self._canceled:
             self._downloading = False
@@ -160,6 +158,7 @@ class _UpdateAutoSuccess(Service):
         # Continue sending more status.
         return True
 
+    @log_and_exit
     @method('com.canonical.SystemImage', out_signature='s')
     def PauseDownload(self):
         if self._downloading:
@@ -168,6 +167,7 @@ class _UpdateAutoSuccess(Service):
         # Otherwise it's a no-op.
         return ''
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def DownloadUpdate(self):
         self._paused = False
@@ -177,6 +177,7 @@ class _UpdateAutoSuccess(Service):
                 self.UpdateProgress(0, 50.0)
                 GLib.timeout_add(500, self._send_more_status)
 
+    @log_and_exit
     @method('com.canonical.SystemImage', out_signature='s')
     def CancelUpdate(self):
         if self._downloading:
@@ -184,6 +185,7 @@ class _UpdateAutoSuccess(Service):
         # Otherwise it's a no-op.
         return ''
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def ApplyUpdate(self):
         # Always succeeds.
@@ -205,32 +207,33 @@ class _UpdateFailed(Service):
 
     def _reset(self):
         self._failure_count = 1
+        self._last_error = 'mock service failed'
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def Reset(self):
         self._reset()
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def CheckForUpdate(self):
         msg = ('You need some network for downloading'
                if self._failure_count > 0
                else '')
+        # Fake enough of the update status to trick _download() into checking
+        # the failure state.
+        class Update:
+            is_available = True
+        self._update = Update()
         self.UpdateAvailableStatus(
-            True, False, '42', 1337 * 1024 * 1024,
+            True, False, '42', 1337 * MiB,
             '1983-09-13T12:13:14',
-            #[
-            #{'description': 'Ubuntu Edge support',
-            # 'description-en_GB': 'change the background colour',
-            # 'description-fr': "Support d'Ubuntu Edge",
-            #},
-            #{'description':
-            # 'Flipped container with 200% boot speed improvement',
-            #}],
             msg)
         if self._failure_count > 0:
             self._failure_count += 1
             self.UpdateFailed(self._failure_count, msg)
 
+    @log_and_exit
     @method('com.canonical.SystemImage', out_signature='s')
     def CancelUpdate(self):
         self._failure_count = 0
@@ -238,26 +241,21 @@ class _UpdateFailed(Service):
 
 
 class _FailApply(Service):
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def Reset(self):
         pass
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def CheckForUpdate(self):
         self.UpdateAvailableStatus(
-            True, False, '42', 1337 * 1024 * 1024,
+            True, False, '42', 1337 * MiB,
             '1983-09-13T12:13:14',
-            #[
-            #{'description': 'Ubuntu Edge support',
-            # 'description-en_GB': 'change the background colour',
-            # 'description-fr': "Support d'Ubuntu Edge",
-            #},
-            #{'description':
-            # 'Flipped container with 200% boot speed improvement',
-            #}],
             '')
         self.UpdateDownloaded()
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def ApplyUpdate(self):
         # The update cannot be applied.
@@ -267,90 +265,131 @@ class _FailApply(Service):
 
 
 class _FailResume(Service):
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def Reset(self):
         pass
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def CheckForUpdate(self):
         self.UpdateAvailableStatus(
-            True, False, '42', 1337 * 1024 * 1024,
+            True, False, '42', 1337 * MiB,
             '1983-09-13T12:13:14',
-            #[
-            #{'description': 'Ubuntu Edge support',
-            # 'description-en_GB': 'change the background colour',
-            # 'description-fr': "Support d'Ubuntu Edge",
-            #},
-            #{'description':
-            # 'Flipped container with 200% boot speed improvement',
-            #}],
             '')
         self.UpdatePaused(42)
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def DownloadUpdate(self):
         self.UpdateFailed(9, 'You need some network for downloading')
 
 
 class _FailPause(Service):
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def Reset(self):
         pass
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def CheckForUpdate(self):
         self.UpdateAvailableStatus(
-            True, True, '42', 1337 * 1024 * 1024,
+            True, True, '42', 1337 * MiB,
             '1983-09-13T12:13:14',
-            #[
-            #{'description': 'Ubuntu Edge support',
-            # 'description-en_GB': 'change the background colour',
-            # 'description-fr': "Support d'Ubuntu Edge",
-            #},
-            #{'description':
-            # 'Flipped container with 200% boot speed improvement',
-            #}],
             '')
         self.UpdateProgress(10, 0)
 
+    @log_and_exit
     @method('com.canonical.SystemImage', out_signature='s')
     def PauseDownload(self):
         return 'no no, not now'
 
 
 class _NoUpdate(Service):
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def Reset(self):
         pass
 
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def CheckForUpdate(self):
         GLib.timeout_add_seconds(3, self._send_status)
 
+    @log_and_exit
     def _send_status(self):
         self.UpdateAvailableStatus(
             False, False, '', 0,
             '1983-09-13T12:13:14',
-            #[
-            #{'description': 'Ubuntu Edge support',
-            # 'description-en_GB': 'change the background colour',
-            # 'description-fr': "Support d'Ubuntu Edge",
-            #},
-            #{'description':
-            # 'Flipped container with 200% boot speed improvement',
-            #}],
             '')
 
 
 class _MoreInfo(Service):
+    def __init__(self, bus, object_path, loop):
+        super().__init__(bus, object_path, loop)
+        self._buildno = 45
+        self._device = 'nexus11'
+        self._channel = 'daily-proposed'
+        self._updated = '2099-08-01 04:45:45'
+        self._version = 'ubuntu=123,mako=456,custom=789'
+        self._checked = '2099-08-01 04:45:00'
+        self._target = 53
+
+    @log_and_exit
     @method('com.canonical.SystemImage')
     def Reset(self):
         pass
 
+    @log_and_exit
     @method('com.canonical.SystemImage', out_signature='isssa{ss}')
     def Info(self):
-        return (45, 'nexus11', 'daily-proposed', '2099-08-01 04:45:45',
-                dict(ubuntu='123', mako='456', custom='789'))
+        return (self._buildno, self._device, self._channel, self._updated,
+                version_detail(self._version))
+
+    @log_and_exit
+    @method('com.canonical.SystemImage', out_signature='a{ss}')
+    def Information(self):
+        return dict(current_build_number=str(self._buildno),
+                    device_name=self._device,
+                    channel_name=self._channel,
+                    last_update_date=self._updated,
+                    version_detail=self._version,
+                    last_check_date=self._checked,
+                    target_build_number=str(self._target))
+
+
+class _Crasher(Service):
+    @log_and_exit
+    @method('com.canonical.SystemImage')
+    def Crash(self):
+        1/0
+
+    @log_and_exit
+    @signal('com.canonical.SystemImage')
+    def SignalCrash(self):
+        1/0
+
+    @log_and_exit
+    @signal('com.canonical.SystemImage')
+    def SignalOkay(self):
+        pass
+
+    @log_and_exit
+    @method('com.canonical.SystemImage')
+    def CrashSignal(self):
+        self.SignalCrash()
+
+    @log_and_exit
+    @method('com.canonical.SystemImage')
+    def Okay(self):
+        pass
+
+    @log_and_exit
+    @method('com.canonical.SystemImage')
+    def CrashAfterSignal(self):
+        self.SignalOkay()
+        1/0
 
 
 def get_service(testing_mode, system_bus, object_path, loop):
@@ -373,6 +412,8 @@ def get_service(testing_mode, system_bus, object_path, loop):
         ServiceClass = _NoUpdate
     elif testing_mode == 'more-info':
         ServiceClass = _MoreInfo
+    elif testing_mode == 'crasher':
+        ServiceClass = _Crasher
     else:
         raise RuntimeError('Invalid testing mode: {}'.format(testing_mode))
     return ServiceClass(system_bus, object_path, loop)
