@@ -22,6 +22,7 @@ __all__ = [
     'copy',
     'data_path',
     'debug',
+    'debuggable',
     'find_dbus_process',
     'get_channels',
     'get_index',
@@ -32,6 +33,7 @@ __all__ = [
     'setup_keyrings',
     'sign',
     'touch_build',
+    'write_bytes',
     ]
 
 
@@ -48,11 +50,12 @@ import unittest
 from contextlib import ExitStack, contextmanager
 from functools import partial, wraps
 from http.server import HTTPServer, SimpleHTTPRequestHandler
+from pathlib import Path
 from pkg_resources import resource_filename, resource_string as resource_bytes
 from socket import SHUT_RDWR
 from systemimage.channel import Channels
 from systemimage.config import Configuration, config
-from systemimage.helpers import atomic, makedirs, temporary_directory
+from systemimage.helpers import MiB, atomic, makedirs, temporary_directory
 from systemimage.index import Index
 from systemimage.state import State
 from threading import Thread
@@ -219,8 +222,7 @@ def configuration(function):
             with atomic(ini_file) as fp:
                 print(template.format(tmpdir=temp_tmpdir,
                                       vardir=temp_vardir), file=fp)
-            config = Configuration()
-            config.load(ini_file)
+            config = Configuration(ini_file)
             resources.enter_context(
                 patch('systemimage.config._config', config))
             resources.enter_context(
@@ -410,9 +412,31 @@ def touch_build(version, timestamp=None):
             pass
 
 
+def write_bytes(path, size_in_mib):
+    # Write size_in_mib * 1MiB number of bytes to the file in path.
+    with open(path, 'wb') as fp:
+        for chunk in range(size_in_mib):
+            fp.write(b'x' * MiB)
+
+
+def debuggable(fn):
+    def wrapper(*args, **kws):
+        try:
+            path = Path('/tmp/debug.enabled')
+            path.touch()
+            return fn(*args, **kws)
+        finally:
+            path.unlink()
+    return wrapper
+
+
 @contextmanager
-def debug():
-    with open('/tmp/debug.log', 'a', encoding='utf-8') as fp:
+def debug(*, check_flag=False):
+    if not check_flag or os.path.exists('/tmp/debug.enabled'):
+        path = Path('/tmp/debug.log')
+    else:
+        path = Path(os.devnull)
+    with path.open('a', encoding='utf-8') as fp:
         function = partial(print, file=fp)
         function.fp = fp
         yield function
@@ -437,9 +461,10 @@ def find_dbus_process(ini_path):
     # for the specific ini_path for the instance we care about.  Yeah, this
     # all kind of sucks, but should be effective in finding the one we need to
     # track.
+    from systemimage.testing.controller import Controller
     for process in psutil.process_iter():
-        cmdline = SPACE.join(process.cmdline)
-        if 'systemimage.service' in cmdline and ini_path in cmdline:
+        cmdline = SPACE.join(process.cmdline())
+        if Controller.MODULE in cmdline and ini_path in cmdline:
             return process
     return None
 
