@@ -26,9 +26,10 @@ import stat
 import logging
 import unittest
 
+from contextlib import ExitStack, contextmanager
 from datetime import timedelta
 from pkg_resources import resource_filename
-from subprocess import check_output
+from subprocess import CalledProcessError, check_output
 from systemimage.config import Configuration
 from systemimage.device import SystemProperty
 from systemimage.reboot import Reboot
@@ -37,11 +38,25 @@ from systemimage.testing.helpers import configuration, data_path, touch_build
 from unittest.mock import patch
 
 
+@contextmanager
+def _patch_device_hook():
+    # The device hook has two things that generally need patching.  The first
+    # is the logging output, which is just noise for testing purposes, so
+    # silence it.  The second is that the `getprop` command may actually exist
+    # on test system, and we want a consistent environment (i.e. the
+    # assumption that the command does not exist).
+    with ExitStack() as resources:
+        resources.enter_context(patch('systemimage.device.logging.getLogger'))
+        resources.enter_context(
+            patch('systemimage.device.check_output',
+                  side_effect=FileNotFoundError))
+        yield
+
+
 class TestConfiguration(unittest.TestCase):
     def test_defaults(self):
         default_ini = resource_filename('systemimage.data', 'client.ini')
-        config = Configuration()
-        config.load(default_ini)
+        config = Configuration(default_ini)
         # [service]
         self.assertEqual(config.service.base, 'system-image.ubuntu.com')
         self.assertEqual(config.service.http_base,
@@ -55,7 +70,8 @@ class TestConfiguration(unittest.TestCase):
         self.assertEqual(config.system.build_file, '/etc/ubuntu-build')
         self.assertEqual(config.system.logfile,
                          '/var/log/system-image/client.log')
-        self.assertEqual(config.system.loglevel, logging.INFO)
+        self.assertEqual(config.system.loglevel,
+                         (logging.INFO, logging.ERROR))
         self.assertEqual(config.system.settings_db,
                          '/var/lib/system-image/settings.db')
         # [hooks]
@@ -86,8 +102,7 @@ class TestConfiguration(unittest.TestCase):
         # Read a basic .ini file and check that the various attributes and
         # values are correct.
         ini_file = data_path('config_01.ini')
-        config = Configuration()
-        config.load(ini_file)
+        config = Configuration(ini_file)
         # [service]
         self.assertEqual(config.service.base, 'phablet.example.com')
         self.assertEqual(config.service.http_base,
@@ -101,7 +116,8 @@ class TestConfiguration(unittest.TestCase):
         self.assertEqual(config.system.build_file, '/etc/ubuntu-build')
         self.assertEqual(config.system.logfile,
                          '/var/log/system-image/client.log')
-        self.assertEqual(config.system.loglevel, logging.ERROR)
+        self.assertEqual(config.system.loglevel,
+                         (logging.ERROR, logging.ERROR))
         self.assertEqual(config.system.settings_db,
                          '/var/lib/phablet/settings.db')
         self.assertEqual(config.system.timeout, timedelta(seconds=10))
@@ -125,11 +141,18 @@ class TestConfiguration(unittest.TestCase):
         # [dbus]
         self.assertEqual(config.dbus.lifetime.total_seconds(), 120)
 
+    def test_special_dbus_logging_level(self):
+        # Read a config.ini that has a loglevel value with an explicit dbus
+        # logging level.
+        ini_file = data_path('config_10.ini')
+        config = Configuration(ini_file)
+        self.assertEqual(config.system.loglevel,
+                         (logging.CRITICAL, logging.DEBUG))
+
     def test_nonstandard_ports(self):
         # config_02.ini has non-standard http and https ports.
         ini_file = data_path('config_02.ini')
-        config = Configuration()
-        config.load(ini_file)
+        config = Configuration(ini_file)
         self.assertEqual(config.service.base, 'phablet.example.com')
         self.assertEqual(config.service.http_base,
                          'http://phablet.example.com:8080')
@@ -139,8 +162,7 @@ class TestConfiguration(unittest.TestCase):
     def test_disabled_http_port(self):
         # config_05.ini has http port disabled and non-standard https port.
         ini_file = data_path('config_05.ini')
-        config = Configuration()
-        config.load(ini_file)
+        config = Configuration(ini_file)
         self.assertEqual(config.service.base, 'phablet.example.com')
         self.assertEqual(config.service.http_base,
                          'https://phablet.example.com:80443')
@@ -150,8 +172,7 @@ class TestConfiguration(unittest.TestCase):
     def test_disabled_https_port(self):
         # config_06.ini has https port disabled and standard http port.
         ini_file = data_path('config_06.ini')
-        config = Configuration()
-        config.load(ini_file)
+        config = Configuration(ini_file)
         self.assertEqual(config.service.base, 'phablet.example.com')
         self.assertEqual(config.service.http_base,
                          'http://phablet.example.com')
@@ -179,22 +200,19 @@ class TestConfiguration(unittest.TestCase):
     def test_get_build_number(self, ini_file):
         # The current build number is stored in a file specified in the
         # configuration file.
-        config = Configuration()
-        config.load(ini_file)
+        config = Configuration(ini_file)
         touch_build(1500)
         self.assertEqual(config.build_number, 1500)
 
     @configuration
     def test_get_build_number_missing(self, ini_file):
         # The build file is missing, so the build number defaults to 0.
-        config = Configuration()
-        config.load(ini_file)
+        config = Configuration(ini_file)
         self.assertEqual(config.build_number, 0)
 
     @configuration
     def test_get_device_name(self, ini_file):
-        config = Configuration()
-        config.load(ini_file)
+        config = Configuration(ini_file)
         # The device name as we'd expect it to work on a real image.
         with patch('systemimage.device.check_output', return_value='nexus7'):
             self.assertEqual(config.device, 'nexus7')
@@ -206,18 +224,25 @@ class TestConfiguration(unittest.TestCase):
         config = Configuration()
         # Silence the log exceptions this will provoke.
         with patch('systemimage.device.logging.getLogger'):
+            # It's possible getprop actually does exist on the system.
+            with patch('systemimage.device.check_output',
+                       side_effect=CalledProcessError(1, 'ignore')):
+                self.assertEqual(config.device, '?')
+
+    def test_device_no_getprop_fallback(self):
+        # Like above, but a FileNotFoundError occurs instead.
+        config = Configuration()
+        with _patch_device_hook():
             self.assertEqual(config.device, '?')
 
     @configuration
     def test_get_channel(self, ini_file):
-        config = Configuration()
-        config.load(ini_file)
+        config = Configuration(ini_file)
         self.assertEqual(config.channel, 'stable')
 
     @configuration
     def test_overrides(self, ini_file):
-        config = Configuration()
-        config.load(ini_file)
+        config = Configuration(ini_file)
         self.assertEqual(config.build_number, 0)
         self.assertEqual(config.device, 'nexus7')
         self.assertEqual(config.channel, 'stable')
@@ -250,8 +275,7 @@ class TestConfiguration(unittest.TestCase):
         # If a /etc/system-image/channels.ini file exists, it overrides any
         # previously set options.
         default_ini = resource_filename('systemimage.data', 'client.ini')
-        config = Configuration()
-        config.load(default_ini)
+        config = Configuration(default_ini)
         # [service]
         self.assertEqual(config.service.base, 'system-image.ubuntu.com')
         self.assertEqual(config.service.http_base,
@@ -275,8 +299,7 @@ class TestConfiguration(unittest.TestCase):
     def test_channel_ini_may_override_system_section(self):
         # Overrides may include the [system] section.
         default_ini = resource_filename('systemimage.data', 'client.ini')
-        config = Configuration()
-        config.load(default_ini)
+        config = Configuration(default_ini)
         channel_ini = data_path('channel_02.ini')
         config.load(channel_ini, override=True)
         self.assertEqual(config.system.build_file,
@@ -288,8 +311,7 @@ class TestConfiguration(unittest.TestCase):
     def test_channel_ini_ignored_sections(self):
         # Only the [service] and [system] section in channel.ini is used.
         default_ini = resource_filename('systemimage.data', 'client.ini')
-        config = Configuration()
-        config.load(default_ini)
+        config = Configuration(default_ini)
         channel_ini = data_path('channel_02.ini')
         config.load(channel_ini, override=True)
         # channel_02.ini sets this to 1h, but it's not overridden.
@@ -298,8 +320,7 @@ class TestConfiguration(unittest.TestCase):
     def test_tempdir(self):
         # config.tempdir is randomly created.
         default_ini = resource_filename('systemimage.data', 'client.ini')
-        config = Configuration()
-        config.load(default_ini)
+        config = Configuration(default_ini)
         self.assertEqual(config.tempdir[:18], '/tmp/system-image-')
         self.assertEqual(stat.filemode(os.stat(config.tempdir).st_mode),
                          'drwx--S---')
@@ -323,3 +344,37 @@ class TestConfiguration(unittest.TestCase):
         stdout = check_output(command, universal_newlines=True)
         self.assertEqual(stdout[:29], 'drwx--S--- /tmp/system-image-')
         self.assertFalse(os.path.exists(stdout.split()[1]))
+
+    def test_constructor(self):
+        # LP: #1342183: Configuration constructor takes an ini_file argument.
+        config = Configuration(data_path('config_01.ini'))
+        self.assertEqual(config.service.base, 'phablet.example.com')
+
+    def test_main_ini_file_must_contain_system_stanza(self):
+        # It's okay if an override is missing the [system] stanza, but the
+        # main ini file (i.e. non-override) must contain it.
+        ini_file = data_path('config_09.ini')
+        config = Configuration()
+        self.assertRaises(KeyError, config.load, ini_file)
+
+    def test_channel_ini_device_override(self):
+        # A channel.ini file can override the device name.
+        config = Configuration(data_path('config_01.ini'))
+        config.load(data_path('channel_06.ini'), override=True)
+        self.assertEqual(config.device, 'shoephone')
+
+    def test_channel_ini_missing_device_override(self):
+        # The channel.ini can omit the [service]device setting, in which case
+        # the hook is still used.
+        config = Configuration(data_path('config_01.ini'))
+        config.load(data_path('channel_05.ini'), override=True)
+        with _patch_device_hook():
+            self.assertEqual(config.device, '?')
+
+    def test_channel_ini_empty_device_override(self):
+        # The channel.ini can have an empty [service]device setting, in which
+        # case the hook is still used.
+        config = Configuration(data_path('config_01.ini'))
+        config.load(data_path('channel_07.ini'), override=True)
+        with _patch_device_hook():
+            self.assertEqual(config.device, '?')
